/*****************************************************************************
 * channel.c: channel for Bt848 frame grabber driver
 *****************************************************************************
 * $Id: channel.c,v 1.26 2003/10/01 06:32:20 pingus77 Exp $
 *****************************************************************************
 * Copyright (C) 2001 Keuleu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 *  Original code:
 *  channel for Bt848 frame grabber driver
 *
 *  Copyright (C) 1996,97 Marcus Metzler (mocm@thp.uni-koeln.de)
 *
 *  many changes by Gerd Knorr <kraxel@cs.tu-berlin.de>
 *      [ hmm, think by now nearly nothing left from the original code ... ]
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <X11/Intrinsic.h>
#ifndef NO_X11
# include <X11/StringDefs.h>
# include <X11/Xaw/XawInit.h>
# include <X11/Xaw/Command.h>
#endif

#include "config.h"
#include "strtab.h"
#include "grab.h"
#include "x11.h"
#include "channels.h"
#include "channel.h"
#include "mixer.h"
#include "plugin.h"
#include "memcpy.h"
#include "deinterlace.h"
#include "divx_ui.h"
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
#include "divx.h"
#endif

/* ----------------------------------------------------------------------- */
/* misc common stuff, not only channel related                             */

struct CHANNEL defaults = {
  "defaults",      // name
  NULL,            // key
  "33",             // cname
  0,               // channel
  0,               // fine
  0,               // freq
  CAPTURE_OVERLAY, // capture
  0,               // source
  2,               // norm
  19000,           // color
  34000,           // bright
  32768,           // hue
  28000,           // contrast
  0,               // pixmap
  0,               // button
  0,               // ckey
  DEINTERLACE_NONE // deinterlace
};

struct CHANNEL **channels = NULL;

int count = 0;
int have_mixer = 0;

int cur_sender = -1, cur_channel = 5, cur_fine = 0;
int cur_norm = -1, cur_input = -1;

int cur_maxwidth = 768, cur_maxheight = 576, cur_secam = 0;
int cur_maxfpsnum = 25, cur_maxfpsden = 1;
int chan_tab = 4;

extern PLUGIN *plugin;
extern char user_dir[255];
extern int have_config;
extern struct GRABBER *grabbers[];
extern int grabber;
extern int fs_width, fs_height;
extern int fs_xoff, fs_yoff;
extern int pix_width, pix_height;
extern int owidth, oheight;
extern int wwidth, wheight;
extern int cur_color, cur_bright, cur_hue, cur_contrast, cur_capture,
  cur_volume;
extern struct STRTAB deinterlace_list[];
extern int deinterlace_method;
extern void set_deinterlace(int j);
#ifdef USE_LIBXOSD
extern char *font, *font_vtx, *colour;
#endif
#ifndef NO_X11
extern Display *dpy;
void button_cb (Widget widget, XtPointer clientdata, XtPointer call_data);

extern Widget app_shell, chan_box, chan_viewport, tv, opt_paned, c_stayontop, c_decoration;
#endif

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
extern divx_t divx;
#endif
extern char *mixer;
extern char *divx_file_path;
extern char *subtitles_pageno_str;
char *divx_path_temp;

extern int memcpy_method;
struct STRTAB memcpy_list[] = {
  {MEMCPY_PROBE,  "probe"},
  {MEMCPY_GLIBC,  "glibc"},
  {MEMCPY_KERNEL, "kernel"},
  {MEMCPY_MMX,    "mmx"},
  {MEMCPY_MMXEXT, "mmxext"},
  {MEMCPY_SSE,    "sse"},
  {-1, NULL}
};

/* ----------------------------------------------------------------------- */
void save_config (void);
/* ----------------------------------------------------------------------- */

int
lookup_channel (char *channel)
{
  int i, nr1, nr2;
  char tag1[5], tag2[5];

  if (isdigit (channel[0]))
    {
      tag1[0] = 0;
      nr1 = atoi (channel);
    }
  else
    {
      sscanf (channel, "%4[A-Za-z]%d", tag1, &nr1);
    }

  for (i = 0; i < CHAN_ENTRIES; i++)
    {
      if (isdigit (tvtuner[i].name[0]))
        {
          tag2[0] = 0;
          nr2 = atoi (tvtuner[i].name);
        }
      else
        {
          sscanf (tvtuner[i].name, "%4[A-Za-z]%d", tag2, &nr2);
        }

      /* First simply test for name equality */
      if (strcmp(channel, tvtuner[i].name) == 0)
        break;

      /* If it doesn't match, try to make a guess */
      if (tag1[0] && tag2[0])
        if (nr1 == nr2 && 0 == strcmp (tag1, tag2))
          break;
      if (!tag1[0] && !tag2[0])
        if (nr1 == nr2)
          break;
    }
  if (i == CHAN_ENTRIES)
    return -1;

  return i;
}

int
get_freq (int i)
{
  if (!tvtuner[i].freq[chan_tab])
    return -1;
  return tvtuner[i].freq[chan_tab] * 16 / 1000;
}

int
cf2freq (char *name, int fine)
{
  int i;

  if (-1 == (i = lookup_channel (name)))
    return -1;
  return get_freq (i) + fine;
}

/* ----------------------------------------------------------------------- */

struct STRTAB captab[] = {
  {CAPTURE_OFF, "off"},
  {CAPTURE_OFF, "no"},
  {CAPTURE_OFF, "false"},
  {CAPTURE_OVERLAY, "on"},
  {CAPTURE_OVERLAY, "yes"},
  {CAPTURE_OVERLAY, "true"},
  {CAPTURE_OVERLAY, "overlay"},
  {CAPTURE_GRABDISPLAY, "grab"},
  {CAPTURE_GRABDISPLAY, "grabdisplay"},
  {-1, NULL}
};

/* just malloc memory for a new channel ... */
struct CHANNEL *
add_channel (char *name)
{
  static int alloc_count = 0;
  struct CHANNEL *channel;

  if (alloc_count == count)
    {
      alloc_count += 16;
      if (alloc_count == 16)
        channels = malloc (sizeof (struct CHANNEL *) * alloc_count);
      else
        channels =
          realloc (channels, sizeof (struct CHANNEL *) * alloc_count);
    }
  channel = channels[count++] = malloc (sizeof (struct CHANNEL));
  memcpy (channel, &defaults, sizeof (struct CHANNEL));
  channel->name = strdup (name);
  return channel;
}

#ifndef NO_X11
void
hotkey_channel (struct CHANNEL *channel)
{
  char str[100], key[32], ctrl[16];

  if (NULL == channel->key)
    return;
  if (2 == sscanf (channel->key, "%15[A-Za-z0-9_]+%31[A-Za-z0-9_]",
                   ctrl, key))
    sprintf (str, "%s<Key>%s: SetChannel(\"%s\")", ctrl, key, channel->name);
  else
    sprintf (str, "<Key>%s: SetChannel(\"%s\")", channel->key, channel->name);
  XtOverrideTranslations (tv, XtParseTranslationTable (str));
  XtOverrideTranslations (opt_paned, XtParseTranslationTable (str));
  XtOverrideTranslations (chan_viewport, XtParseTranslationTable (str));
}
#endif

/* ... and initalize later */
void
configure_channel (struct CHANNEL *channel)
{
  channel->freq = get_freq (channel->channel) + channel->fine;
#ifndef NO_X11
  channel->button =
    XtVaCreateManagedWidget (channel->name,
                             commandWidgetClass, chan_box,
                             XtNwidth, pix_width,
                             XtNheight, pix_height, NULL);
  XtAddCallback (channel->button, XtNcallback, button_cb,
                 (XtPointer *) channel);
  hotkey_channel (channel);
#endif
}

/* delete channel */
void
del_channel (int i)
{
#ifndef NO_X11
  XtDestroyWidget (channels[i]->button);
#endif
  free (channels[i]->name);
  if (channels[i]->key)
    free (channels[i]->key);
  free (channels[i]);
  count--;
  if (i < count)
    memmove (channels + i, channels + i + 1,
             (count - i) * sizeof (struct CHANNEL *));
}


void
update_channel(struct CHANNEL *dest, struct CHANNEL *src)
{
  Pixmap p;
  Widget b;
  int ck;

  p = dest->pixmap;
  b = dest->button;
  ck = dest->ckey;
  memcpy(dest, src, sizeof(struct CHANNEL));
  dest->pixmap = p;
  dest->button = b;
  dest->ckey = ck;
}

void
debug_channel(struct CHANNEL *chan)
{
  printf("\n");
  printf("channel.name        = %s\n", chan->name);
  printf("channel.key         = %s\n", chan->key);
  printf("channel.cname       = %s\n", chan->cname);
  printf("channel.channel     = %d\n", chan->channel);
  printf("channel.fine        = %d\n", chan->fine);
  printf("channel.freq        = %d\n", chan->freq);
  printf("channel.capture     = %s\n", int_to_str(chan->capture, captab));
  printf("channel.source      = %s\n", int_to_str(chan->source, grabbers[grabber]->inputs));
  printf("channel.norm        = %s\n", int_to_str(chan->norm, grabbers[grabber]->norms));
  printf("channel.color       = %d\n", chan->color);
  printf("channel.bright      = %d\n", chan->bright);
  printf("channel.hue         = %d\n", chan->hue);
  printf("channel.contrast    = %d\n", chan->contrast);
  printf("channel.pixmap      = %ld\n", chan->pixmap);
  printf("channel.button      = %p\n", chan->button);
  printf("channel.ckey        = %d\n", chan->ckey);                
  printf("channel.deinterlace = %d\n", chan->deinterlace);                
  printf("\n");
}

/* ----------------------------------------------------------------------- */
int
read_channel(char *filename, char *line, int nr, struct CHANNEL **pt_current, char *tag, char *val, int flag)
{
  char *h;
  int i;
  struct CHANNEL *current = *pt_current;

  if (1 == sscanf (line, "[%99[^]]]", val))
    {
      if (flag == True)
        {
          current = add_channel (val);
          *pt_current = current;
        }
      else
        current->name = strdup(val);

      //continue;
      return 0;
    }

  if (2 != sscanf (line, " %31[^= ] = %99[^\n]", tag, val))
    {
      fprintf (stderr, "%s:%d: parse error\n", filename, nr);
      //continue;
      return 0;
    }

  for (h = val + strlen (val); h > val && *h == ' '; h--)
    *h = '\0';
  
  if (0 == strcmp (tag, "key"))
    {
      current->key = strdup (val);
    }
  else if (0 == strcmp (tag, "capture"))
    {
      if (-1 != (i = str_to_int (val, captab)))
        current->capture = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "source"))
    {
      if (-1 != (i = str_to_int (val, grabbers[grabber]->inputs)))
        current->source = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "norm"))
    {
      if (-1 != (i = str_to_int (val, grabbers[grabber]->norms)))
        current->norm = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else if (0 == strcmp (tag, "channel"))
    {
      current->cname = strdup (val);
      current->channel = lookup_channel (current->cname);
      if (-1 == current->channel)
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  // Patch provided by Henk
  //This makes finetuning of freq. configured channels possible
  else if (0 == strcmp(tag, "freq"))
    {
      int n;
      unsigned d;
  
      current->freq = (int)(atof(val) * 1000);
      n = 0;
      d = (unsigned)-1;
      for (i = 0; i < CHAN_ENTRIES; i++)
        {
          if (abs(tvtuner[i].freq[chan_tab] - current->freq) < d)
            {
              d = abs(tvtuner[i].freq[chan_tab] - current->freq);
              n = i;
            }
        }
      current->cname = strdup(tvtuner[n].name);
      current->channel = lookup_channel(current->cname);
      current->fine = (current->freq - tvtuner[n].freq[chan_tab]) * 16 / 1000;
      current->freq = tvtuner[n].freq[chan_tab] * 16 / 1000;
      // debug_channel(current);
    }
  // End of Patch provided by Henk
  else if (0 == strcmp (tag, "fine"))
    {
      current->fine = atoi (val);
    }
  else if (0 == strcmp (tag, "color"))
    {
        current->color = atoi (val);
    }
  else if (0 == strcmp (tag, "bright"))
    {
      current->bright = atoi (val);
    }
  else if (0 == strcmp (tag, "hue"))
    {
      current->hue = atoi (val);
    }
  else if (0 == strcmp (tag, "contrast"))
    {
      current->contrast = atoi (val);
    }
  else if (0 == strcmp (tag, "deinterlace"))
    {
      if (-1 != (i = str_to_int (val, deinterlace_list)))
        current->deinterlace = i;
      else
        fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                 filename, nr, tag, val);
    }
  else
    {
      // nothing has been find
      return -1;
    }

  // something relevant has been find and hopefully interpreted
  return 0;
}

void
read_config (void)
{
  FILE *fp;
  char filename[100], line[100], tag[32], val[100];
  int i, j, nr = 0;
  float k = 0.0;
  struct CHANNEL *current = &defaults;

  sprintf (filename, "%s/%s", user_dir, "xawdecoderc");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf (stderr, "Can't open config file %s:", filename);
      fprintf (stderr, "writing a default one.\n");
      fprintf (stderr, "Please edit %s to suit your needs\n", filename);
      /* Write a default config file */
      save_config ();

      /* And reread it ! */
      if ((fp = fopen (filename, "r")) == (FILE *) NULL)
        return;
    }
  have_config = 1;

  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;

      if (read_channel(filename, line, nr, &current, tag, val, True) == 0)
        {
          continue;
        }
      else if (0 == count && 0 == strcmp (tag, "mixer"))
        {
          mixer = strdup (val);
          if (-1 != mixer_open ("/dev/mixer", val))
            have_mixer = 1;
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "freqtab"))
        {
          if (-1 != (i = str_to_int (val, chan_names)))
            chan_tab = i;
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "fullscreen"))
        {
          if (2 != sscanf (val, "%d x %d", &fs_width, &fs_height))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              fs_width = fs_height = 0;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "windowsize"))
        {
          if (2 == sscanf (val, "%d x %d", &i, &j))
            {
              video_set_size (i, j, ADAPT_SIZE);
            }
          else
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
      else if (0 == count && 0 == strcmp (tag, "pixsize"))
        {
          if (2 != sscanf (val, "%d x %d", &pix_width, &pix_height))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              pix_width = 128;
              pix_height = 96;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "wm-off-by"))
        {
          if (2 != sscanf (val, "%d %d", &fs_xoff, &fs_yoff))
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
              fs_xoff = fs_yoff = 0;
            }
        }
      else if (0 == count && 0 == strcmp (tag, "colorkey"))
        {
          if (1 == sscanf (val, "%d", &i))
            {
              set_colorkey(dpy, i);
            }
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "subpageno"))
	 subtitles_pageno_str = strdup(val);

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
      else if (0 == count && 0 == strcmp (tag, "codec"))
        {
	  if(0) {}
#ifdef HAVE_DIVX4LINUX
          else if (0 == strcmp(val,"DivX"))
            {
              divx.codec = DIVX4LINUX;
            }
#endif
#ifdef HAVE_FFMPEG
          else if (0 == strcmp(val,"FFMPEG - Mpeg4"))
            {
              divx.codec = FFMPEG_MPEG4;
            }
          else if (0 == strcmp(val,"FFMPEG - Mpeg1Video"))
            {
              divx.codec = FFMPEG_MPEG1;
            }
#endif
#ifdef HAVE_XVID
          else if (0 == strcmp(val,"XviD"))
            {
              divx.codec = XVID;
            }
#endif
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "width"))
        {
          if (1 != sscanf(val,"%d",&divx.width))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "height"))
        {
          if (1 != sscanf(val,"%d",&divx.height))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "bitrate"))
        {
          if (1 != sscanf(val,"%d",&i))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.bitrate = i*1000;
	  }
        }
     else if (0 == count && 0 == strcmp (tag, "quality"))
       {
         switch(divx.codec)
	   {
	 #ifdef HAVE_DIVX4LINUX
	   case DIVX4LINUX:
	     if ( 0 == strcmp (val, "Fastest"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "Fast"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "Slow"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Slowest"))
	       divx.quality = 5;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
	 #ifdef HAVE_FFMPEG
	   case FFMPEG_MPEG4:
	   case FFMPEG_MPEG1:
	     if ( 0 == strcmp (val, "No ME"))
	       divx.quality = 0;
	     else if ( 0 == strcmp (val, "Very Low"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "Low"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "High"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Very High"))
	       divx.quality = 5;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
	 #ifdef HAVE_XVID
	   case XVID:
	     if ( 0 == strcmp (val, "H623, No ME"))
	       divx.quality = 0;
	     else if ( 0 == strcmp (val, "MPEG, Low"))
	       divx.quality = 1;
	     else if ( 0 == strcmp (val, "H623, Low"))
	       divx.quality = 2;
	     else if ( 0 == strcmp (val, "H623, Medium"))
	       divx.quality = 3;
	     else if ( 0 == strcmp (val, "H623, High"))
	       divx.quality = 4;
	     else if ( 0 == strcmp (val, "Very High"))
	       divx.quality = 5;
	     else if ( 0 == strcmp (val, "Ultra High"))
	       divx.quality = 6;
	     else
               fprintf (stderr, "%s:%d: invalid value for %s: %s\n", filename, nr, tag, val);
	     break;
	 #endif
           }
	}
      else if (0 == count && 0 == strcmp (tag, "stereo_mode"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.stereo = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.stereo = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "compress_audio"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.compress_audio = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.compress_audio = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_bitrate"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_bitrate))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_quality"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_quality))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_vbr_mode"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.mp3_vbr = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.mp3_vbr = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "mp3_vbr_quality"))
        {
          if (1 != sscanf(val,"%d",&divx.mp3_vbr_quality))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "fps"))
        {
          if (1 != sscanf(val,"%d",&i))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.fps = i*1000;
	  }
        }
      else if (0 == count && 0 == strcmp (tag, "min_quantizer"))
        {
          if (1 != sscanf(val,"%d",&divx.min_quantizer))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "max_quantizer"))
        {
          if (1 != sscanf(val,"%d",&divx.max_quantizer))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "audio_buffer_size"))
        {
          if (1 != sscanf(val,"%d",&divx.audio_buffer_size))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
        }
      else if (0 == count && 0 == strcmp (tag, "max_gap"))
        {
          if (1 != sscanf(val,"%f",&k))
            {
!            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
	    }
	  else
	  {
	    divx.maxgap = k/1000.0;
	  }
        }
      else if (0 == count && 0 == strcmp (tag, "display_frame"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.display_frames = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.display_frames = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "record_sub"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.sub = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.sub = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      else if (0 == count && 0 == strcmp (tag, "record_chg"))
        {
         if ( 0 == strcmp (val, "off") )
	   divx.chg = 0;
	 else if ( 0 == strcmp (val, "on") )
	   divx.chg = 1;
	 else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      /* read divx file path into xawdecoderc and give it to DivX GUI */
      else if (0 == count && 0 == strcmp (tag, "divx_path"))
	 divx_file_path = strdup(val);

#endif
#ifdef USE_LIBXOSD
      else if (0 == count && 0 == strcmp (tag, "xosd_color"))
        {
		colour = strdup(val);
        }
      else if (0 == count && 0 == strcmp (tag, "xosd_font"))
        {
		font = strdup(val);
        }
       else if (0 == count && 0 == strcmp (tag, "xosd_vtxfont"))
        {
		font_vtx = strdup(val);
        }
#endif
      /*
      else if (0 == count && 0 == strcmp (tag, "deinterlace"))
        {
          if (-1 != (i = str_to_int (val, deinterlace_list)))
            {
              set_deinterlace(i);
            }
          else
            fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                     filename, nr, tag, val);
        }
      */
      else
        {
          int ret = -1;
          if (plugin != NULL)
            ret = plugin->read_config(&count, tag, val, nr, filename);
          if (ret == -1)
            fprintf (stderr, "%s:%d: unknown tag %s\n", filename, nr, tag);
        }
    }
  fclose (fp);

  /* calculate channel frequencies */
  defaults.channel = lookup_channel (defaults.cname);
  defaults.freq = get_freq (defaults.channel) + defaults.fine;
  for (i = 0; i < count; i++)
    configure_channel (channels[i]);
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
  create_divxwin();
#endif
}

/* ----------------------------------------------------------------------- */

void
save_channel(FILE *fp, int i)
{
  fprintf (fp, "[%s]\n", channels[i]->name);

  fprintf (fp, "channel = %s\n", tvtuner[channels[i]->channel].name);

  if (defaults.fine != channels[i]->fine)
    fprintf (fp, "fine = %d\n", channels[i]->fine);

  if (defaults.norm != channels[i]->norm)
    fprintf (fp, "norm = %s\n",
             int_to_str (channels[i]->norm, grabbers[grabber]->norms));

  if (channels[i]->key != NULL)
    fprintf (fp, "key = %s\n", channels[i]->key);

  if (defaults.capture != channels[i]->capture)
    fprintf (fp, "capture = %s\n",
             int_to_str (channels[i]->capture, captab));

  if (defaults.source != channels[i]->source)
    fprintf (fp, "source = %s\n",
             int_to_str (channels[i]->source, grabbers[grabber]->inputs));
  
  if (defaults.color != channels[i]->color)
    fprintf (fp, "color = %d\n", channels[i]->color);

  if (defaults.bright != channels[i]->bright)
    fprintf (fp, "bright = %d\n", channels[i]->bright);

  if (defaults.hue != channels[i]->hue)
    fprintf (fp, "hue = %d\n", channels[i]->hue);

  if (defaults.contrast != channels[i]->contrast)
    fprintf (fp, "contrast = %d\n", channels[i]->contrast);
  
  if (defaults.deinterlace != channels[i]->deinterlace)
    fprintf (fp, "deinterlace = %s\n", int_to_str(channels[i]->deinterlace, deinterlace_list));

  fprintf (fp, "\n");
}


#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
void
save_record(FILE *fp)
{
  fprintf (fp, "#\n# record options\n#\n");
  switch(divx.codec)
  {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      fprintf (fp, "codec = DivX\n");
      break;
#endif

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
      fprintf (fp, "codec = FFMPEG - Mpeg4\n");
      break;
    case FFMPEG_MPEG1:
      fprintf (fp, "codec = FFMPEG - Mpeg1Video\n");
      break;
#endif

#ifdef HAVE_XVID
    case XVID:
      fprintf (fp, "codec = XviD\n");
      break;
#endif
   }
   fprintf (fp,"width = %d\n",divx.width);
   fprintf (fp,"height = %d\n",divx.height);
   fprintf (fp,"bitrate = %ld\n",divx.bitrate/1000);
  switch(divx.codec)
  {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      switch(divx.quality)
      {
      	case 1:
      	 fprintf (fp,"quality = Fastest\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = Fast\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = Slow\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Slowest\n");
      	 break;
      }
      break;
#endif

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
    case FFMPEG_MPEG1:
      switch(divx.quality)
      {
      	case 0:
      	 fprintf (fp,"quality = No ME\n");
      	 break;
      	case 1:
      	 fprintf (fp,"quality = Very Low\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = Low\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = High\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Very High\n");
      	 break;
      }
      break;
#endif

#ifdef HAVE_XVID
    case XVID:
      switch(divx.quality)
      {
      	case 0:
      	 fprintf (fp,"quality = H623, No ME\n");
      	 break;
      	case 1:
      	 fprintf (fp,"quality = MPEG, Low\n");
      	 break;
      	case 2:
      	 fprintf (fp,"quality = H623, Low\n");
      	 break;
      	case 3:
      	 fprintf (fp,"quality = H623, Medium\n");
      	 break;
      	case 4:
      	 fprintf (fp,"quality = H623, High\n");
      	 break;
      	case 5:
      	 fprintf (fp,"quality = Very High\n");
      	 break;
      	case 6:
      	 fprintf (fp,"quality = Ultra High\n");
      	 break;
      }
      break;
#endif
   }
   switch(divx.stereo)
   {
     case 0:
       fprintf (fp,"stereo_mode = off\n");
       break;
     case 1:
       fprintf (fp,"stereo_mode = on\n");
       break;
   }
   switch(divx.compress_audio)
   {
     case 0:
       fprintf (fp,"compress_audio = off\n");
       break;
     case 1:
       fprintf (fp,"compress_audio = on\n");
       break;
   }
   fprintf (fp,"mp3_bitrate = %d\n",divx.mp3_bitrate);
   fprintf (fp,"mp3_quality = %d\n",divx.mp3_quality);
   switch(divx.mp3_vbr)
   {
     case 0:
       fprintf (fp,"mp3_vbr_mode = off\n");
       break;
     case 1:
       fprintf (fp,"mp3_vbr_mode = on\n");
       break;
   }
   fprintf (fp,"mp3_vbr_quality = %d\n",divx.mp3_vbr_quality);
   fprintf (fp,"fps = %d\n",divx.fps/1000);
   fprintf (fp,"max_gap = %f\n",divx.maxgap*1000.0);
   fprintf (fp,"min_quantizer = %d\n",divx.min_quantizer);
   fprintf (fp,"max_quantizer = %d\n",divx.max_quantizer);
   fprintf (fp,"audio_buffer_size = %d\n",divx.audio_buffer_size);

/* write divx file path into xawdecoderc */
   get_divxfilepath(&divx_path_temp);
   fprintf (fp,"divx_path = %s\n",divx_path_temp);

   switch(divx.display_frames)
   {
     case 0:
       fprintf (fp,"display_frame = off\n");
       break;
     case 1:
       fprintf (fp,"display_frame = on\n");
       break;
   }
   switch(divx.sub)
   {
     case 0:
       fprintf (fp,"record_sub = off\n");
       break;
     case 1:
       fprintf (fp,"record_sub = on\n");
       break;
   }
   switch(divx.chg)
   {
     case 0:
       fprintf (fp,"record_chg = off\n");
       break;
     case 1:
       fprintf (fp,"record_chg = on\n");
       break;
   }
   fprintf(fp,"\n");


}
#endif

void
save_config (void)
{
  char filename1[100], filename2[100];
  FILE *fp;
  int i;
  Dimension w = 0;
  Dimension h = 0;

  sprintf (filename1, "%s/%s", user_dir, "xawdecoderc");
  sprintf (filename2, "%s/%s", user_dir, "xawdecoderc.bak");

  /* delete old backup */
  unlink (filename2);

  /* current becomes backup */
  if (0 == link (filename1, filename2))
    unlink (filename1);

  /* write new one... */
  fp = fopen (filename1, "w");
  if (NULL == fp)
    {
      fprintf (stderr, "can't open config file %s\n", filename1);
      return;
    }

  /* write defaults */
  fprintf(fp, "#\n# Global options\n#\n");

  if (defaults.norm == -1)
    defaults.norm = str_to_int ("SECAM",  grabbers[grabber]->norms);
  fprintf (fp, "norm = %s\n",
           int_to_str (defaults.norm, grabbers[grabber]->norms));

  if ((defaults.capture == -1) || (defaults.capture == str_to_int ("off", captab)))
    defaults.capture = str_to_int ("grabdisplay", captab);
  fprintf (fp, "capture = %s\n", int_to_str (defaults.capture, captab));
/*

// by default it's PAL and the choose could be written now
  if (defaults.norm == -1)
    	defaults.norm = str_to_int ("PAL",  grabbers[grabber]->norms);
  fprintf (fp, "norm = %s\n", int_to_str (cur_norm, grabbers[grabber]->norms));

  if (cur_capture == str_to_int ("off", captab))
    cur_capture = str_to_int ("grab", captab);
    fprintf (fp, "capture = %s\n", int_to_str (cur_capture, captab));
*/
  if (defaults.source == -1)
    defaults.source = str_to_int ("Television", grabbers[grabber]->inputs);
  fprintf (fp, "source = %s\n",
           int_to_str (defaults.source, grabbers[grabber]->inputs));

  if ((defaults.color != 19000) && (defaults.color != 0))
    fprintf (fp, "color = %d\n", defaults.color);

  if ((defaults.bright != 34000) && (defaults.bright != 0))
    fprintf (fp, "bright = %d\n", defaults.bright);

  if ((defaults.hue != 32768) && (defaults.hue != 0))
    fprintf (fp, "hue = %d\n", defaults.hue);

  if ((defaults.contrast != 28000) && (defaults.contrast != 0))
    fprintf (fp, "contrast = %d\n", defaults.contrast);

  fprintf (fp, "\n");

  if (fs_width && fs_height)
    fprintf (fp, "fullscreen = %d x %d\n", fs_width, fs_height);
  else
    fprintf (fp, "fullscreen = 768 x 576\n");

  if (fs_xoff || fs_yoff)
    fprintf (fp, "wm-off-by = %+d%+d\n", fs_xoff, fs_yoff);

  fprintf (fp, "freqtab = %s\n", int_to_str (chan_tab, chan_names));
  fprintf (fp, "pixsize = %d x %d\n", pix_width, pix_height);

  if (fs_width && fs_height)
  {
    XtVaGetValues (tv, XtNwidth, &w, XtNheight, &h, NULL);
    fprintf (fp, "windowsize = %d x %d\n", w, h);
  }
  else fprintf (fp, "windowsize = 384 x 288\n");


  if (mixer)
    fprintf (fp, "mixer = %s\n", mixer);

  if ((i = get_colorkey(dpy)) != -1)
    fprintf (fp, "colorkey = %d\n", i);

  //fprintf (fp, "nbufs = %d\n", nbufs_default);
  if (subtitles_pageno_str == NULL)
  fprintf(fp, "subpageno = 888\n");
  else
  fprintf(fp, "subpageno = %s\n",subtitles_pageno_str);

#ifdef USE_LIBXOSD
  if (colour)
    fprintf (fp, "xosd_color = %s\n", colour);

  if (font)
    fprintf (fp, "xosd_font = %s\n", font);

  if (font_vtx)
    fprintf (fp, "xosd_vtxfont = %s\n", font_vtx);
#endif

  fprintf (fp, "\n");
  /*
  fprintf(fp, "# deinterlace: none | bob | weave | greedy | onefield\n"
              "deinterlace = %s\n", int_to_str(deinterlace_method, deinterlace_list));
  */

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
  save_record(fp);
#endif

  if (plugin != NULL)
    plugin->save_config(fp);

  /* write channels */
  if (count > 0)
    {
      fprintf(fp, "#\n# Channel options\n#\n");
      for (i = 0; i < count; i++)
        {
          save_channel(fp, i);
        }
    }
  else
    {
      fprintf (fp,
               "# For each channel define at least following information\n");
      fprintf (fp, "# [Channel Name]\n");
      fprintf (fp, "# channel = Channel Number\n");
      fprintf (fp, "# norm = PAL|NTSC|SECAM|AUTO\n");
      fprintf (fp,
               "# key = key to press to switch the channel on (KP_End, KP_Next,...)\n");
      fprintf (fp, "# capture = off|grabdisplay|overlay\n");
      fprintf (fp, "\n");
    }



  fclose (fp);
}

int
load_last_channel(struct CHANNEL *last_channel)
{
  FILE *fp;
  char filename[100], line[100], tag[32], val[100];
  int nr = 0;
  int r, status = 0;
  int w, h;

  sprintf (filename, "%s/%s", user_dir, "last_channel");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf(stderr, "load_last_channel: can't open %s\n", filename);
      return -1;
    }

  memcpy(last_channel, &defaults, sizeof(struct CHANNEL));

  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;


      if ((r=read_channel(filename, line, nr, &last_channel, tag, val, False)) == 0)
        {
          continue;
        }
      else if (0 == strcmp (tag, "windowsize"))
        {
          if (2 == sscanf (val, "%d x %d", &w, &h))
            {
              video_set_size (w, h, ADAPT_SIZE);
            }
          else
            {
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
      else
        {
          // Nothing was found;
          status += r;
        }
    }

  if (nr == 0)
    return -1; // file was empty !!!
  else if (status == 0)
    {
      /* calculate channel frequencies */
      last_channel->freq = get_freq (last_channel->channel) + last_channel->fine;
      hotkey_channel (last_channel);
      return 0; // everything went fine
    }
  else
    return -1;
}

void
save_last_channel(int last_channel)
{
  FILE *fp;
  char filename[100];
  //Dimension w, h;
  
  sprintf (filename, "%s/%s", user_dir, "last_channel");
  fp = fopen (filename, "w");
  if (NULL == fp)
    {
      fprintf(stderr, "save_last_channel: can't open %s\n", filename);
      return;
    }

  // Causes more troubles than real added value
  //XtVaGetValues (tv, XtNwidth, &w, XtNheight, &h, NULL);
  //fprintf (fp, "windowsize = %d x %d\n", w, h);

  save_channel(fp, last_channel);

  fclose(fp);
}


void
save_memcpy_method(int method)
{
  FILE *fp;
  char filename[100];

  sprintf (filename, "%s/%s", user_dir, "memcpy_method");
  fp = fopen (filename, "w");

  fprintf(fp, "# memcpy_method: probe | glibc | kernel | mmx | mmxext | sse\n"
              "memcpy_method = %s\n", int_to_str(method, memcpy_list));

  fclose(fp);
}


void
load_memcpy_method(void)
{
  FILE *fp;
  char filename[100];
  char line[100];
  int nr;
  char *h;
  char tag[32], val[100];
  int i;

  sprintf (filename, "%s/%s", user_dir, "memcpy_method");
  fp = fopen (filename, "r");
  if (NULL == fp)
    {
      fprintf(stderr, "load_memcpy_method: can't open %s\n", filename);
      return;
    }

  nr = 0;
  while (NULL != fgets (line, 99, fp))
    {
      nr++;
      if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
        continue;

      if (2 != sscanf (line, " %31[^= ] = %99[^\n]", tag, val))
        {
          fprintf (stderr, "%s:%d: parse error\n", filename, nr);
          continue;
        }

      for (h = val + strlen (val); h > val && *h == ' '; h--)
        *h = '\0';
      

      if (0 == strcmp (tag, "memcpy_method"))
        {
          if (-1 != (i = str_to_int (val, memcpy_list)))
            {
              memcpy_method = i;
            }
          else
            {
              memcpy_method = MEMCPY_PROBE;
              fprintf (stderr, "%s:%d: invalid value for %s: %s\n",
                       filename, nr, tag, val);
            }
        }
    }
}

