/*****************************************************************************
 * divx.c: DivX recording
 *****************************************************************************
 * $Id: divx.c,v 1.42 2003/09/20 09:16:20 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2001 Keuleu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
# include <linux/soundcard.h>
# include <sys/ioctl.h>
#include <sys/time.h>
#include "config.h"
#include "writefile.h"

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)

#ifdef HAVE_DIVX4LINUX
//# include "encore2.h"
# include <encore2.h>
#endif

#ifdef HAVE_FFMPEG
# include <ffmpeg/avcodec.h>
#if LIBAVCODEC_BUILD < 4641
#  error we dont support libavcodec prior to build 4641, get the latest libavcodec CVS
#endif
#endif

#ifdef HAVE_XVID
# include <xvid.h>
#endif

#include "divx.h"
#include "avilib.h"

#include <X11/Intrinsic.h>
#include "strtab.h"
#include "channel.h"
#include "plugin.h"
#include "memcpy.h"
#include "grab.h"
#include "divx_ui.h"

#undef  NOAUDIO
#ifndef NOAUDIO
#define SYNCHRO
#endif

void video_get_size(int *w, int *h);
void video_set_size(int w, int h, int force);
void video_get_capture_size(int *w, int *h);
void video_set_capture_size(int w, int h);
void video_lock_size(void);
void video_unlock_size(void);

divx_t divx = {
#if defined (HAVE_XVID)
 XVID,         // codec to use
#elif defined (HAVE_FFMPEG)
 FFMPEG_MPEG4, // codec to use
#else
 DIVX4LINUX,   // codec to use
#endif
      0, // pixmap
   NULL, // convert
    384, // width
    288, // height
 800000, // bitrate (bits/second)
      1, // quality
      2, // min_quantizer
      8, // max_quantizer
  4*441, // audio_buffer_size
      0, // stereo
#ifdef HAVE_LAME
      1, // compress_audio
#else
      0, // compress_audio
#endif
    128, // mp3_bitrate
      5, // mp3_quality
      0, // mp3_vbr
      8, // mp3_vbr_quality
  25000, // fps * 1000
   0.08, // maxgap between audio and video in seconds.
   0.00, // delay between audio and video
      1, // display_frames
      0, // sub
      0, // chg
};
static int       slw; //status of the last AVI_write
static int displaynextframe = 1;

extern int debug;
extern PLUGIN *plugin;

#ifndef NOAUDIO
/*****************************************************************************
 *                       RAW AUDIO DEFINITIONS
 ****************************************************************************/
#define DEC_FREQ    44100
#define FORMAT16    AFMT_S16_LE
#define FORMAT8     AFMT_U8
#define FRAGMENTS   0x0030000b    /* Frags of 2048 bytes */

static int       audiofd = -1;
static char *    device = "/dev/dsp";
static pthread_t audio_thread;
static int       achans;  // nb of audio channels
static int       aformat = 0; // Audio format
static unsigned long audio_frames, audio_buffers;
static int       fps_num, fps_den;

static void *    audio_record(void *);

/*****************************************************************************
 *                       MP3 AUDIO DEFINITIONS
 ****************************************************************************/
#ifdef HAVE_LAME

#include "lame/lame.h"

static lame_global_flags *gfp = NULL;
static char *    lame_buffer = NULL;
static int       lame_buffer_size = 0;

inline static int       lame_encode(short int *raw_buffer, int samples_per_channel);
static void      lame_stop(void);

#endif // HAVE_LAME

#ifdef SYNCHRO
static long long nbsamplesaudios;
static unsigned long dropped_audio_buffers, dropped_video_frames;
static long long dropped_audio_samples;
extern double videostampmin,videostampmax;
double audiostamp=0.0;
#endif // SYNCHRO
#endif // NOAUDIO


/*****************************************************************************
 *                       DIVX4 VIDEO DEFINITIONS
 ****************************************************************************/
static void *    handle = NULL;

#ifdef HAVE_DIVX4LINUX
# if ENCORE_VERSION >= 20021024
   static DivXBitmapInfoHeader *format =NULL;
   static SETTINGS *settings = NULL;
# else
   static ENC_PARAM param;
# endif
static ENC_FRAME frame;
#endif

#ifdef HAVE_FFMPEG
static AVCodec *codec = NULL;
static AVCodecContext *codec_context = NULL;
static AVFrame *picture_420p = NULL;
#endif

#ifdef HAVE_XVID
static XVID_ENC_FRAME  xframe;
static XVID_ENC_PARAM  xparam;
#define FRAMERATE_INCR 1001
static int const xmotion_presets[7] = {
        0,                                                        // Q 0
        PMV_EARLYSTOP16,                                          // Q 1
        PMV_EARLYSTOP16,                                          // Q 2
        PMV_EARLYSTOP16 | PMV_HALFPELREFINE16,                    // Q 3
        PMV_EARLYSTOP16 | PMV_HALFPELREFINE16,                    // Q 4
        PMV_EARLYSTOP16 | PMV_HALFPELREFINE16 | PMV_EARLYSTOP8 |  // Q 5
        PMV_HALFPELREFINE8,
        PMV_EARLYSTOP16 | PMV_HALFPELREFINE16 | PMV_EXTSEARCH16 | // Q 6
        PMV_USESQUARES16 | PMV_EARLYSTOP8 | PMV_HALFPELREFINE8
};
static int const xgeneral_presets[7] = {
        XVID_H263QUANT,                               // Q 0
        XVID_MPEGQUANT,                               // Q 1
        XVID_H263QUANT,                               // Q 2
        XVID_H263QUANT | XVID_HALFPEL,                // Q 3
        XVID_H263QUANT | XVID_HALFPEL | XVID_INTER4V, // Q 4
        XVID_H263QUANT | XVID_HALFPEL | XVID_INTER4V, // Q 5
        XVID_H263QUANT | XVID_HALFPEL | XVID_INTER4V  // Q 6
};
#endif

static unsigned char * convert_buf = NULL;

  // Doc says max encoded frame size is around 6 bytes per pixel...
  // I like the "around" ....
#define ENC_FRAME_SIZE (6*768*576)
static unsigned char * enc_frame = NULL;

static avi_t *   avifile = NULL;
static FILE *    subfile = NULL;

static int       prev_w = 0;
static int       prev_h = 0;
static int       prev_w_capture = 0;
static int       prev_h_capture = 0;
static unsigned long video_frames;
static unsigned long lastsubend;
static struct timeval tv_start;

extern int       AVI_errno;
static int       divx_in_progress = 0;

/*****************************************************************************
 *                      SOME COLOR SPACE CONVERSIONS
 ****************************************************************************/
static void
yuv422_to_yuv420p(unsigned char *d, unsigned char *foo1, unsigned char *foo2,
                  unsigned char *s, int w, int h);

static void
rgb32_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

static void
rgb24_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

static void
rgb16_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

static void
rgb15_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

static void
rgb32_top_down(unsigned char *dest, unsigned char *foo1, unsigned char *foo2,
               unsigned char *src, int width, int height);

static void
gray_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

extern unsigned short hi240_to_rgb[][3];

static void
hi240_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height);

static void
hi240_to_rgb32(unsigned char *lum, unsigned char *cb, unsigned char *cr,
	       unsigned char *src, int width, int height);


/*****************************************************************************
 *                       DIVX VIDEO FUNCTIONS
 ****************************************************************************/
void
divx_init(char *divx_file, int pixmap, int width, int height)
{
#ifdef HAVE_DIVX4LINUX
# if defined(ENC_OPT_VERSION) && defined (ENCORE_VERSION)
   int version_binary;
# endif
#endif
#ifdef HAVE_XVID
  XVID_INIT_PARAM xinit;
#endif
#if defined (HAVE_DIVX4LINUX) || defined (HAVE_XVID)
  int ret;
#endif
  struct timezone tz;
  time_t t;
#ifndef SYNCHRO
  int fps = (double)cur_maxfpsnum * 1000 / cur_maxfpsden + 0.5;
#else
  int fps = divx.fps;
#endif
  char *fourcc = NULL;
  char *filename = NULL;
  char *divx_path = NULL;
  char *codec_name = NULL;
  struct timeval tv;

  if (handle != NULL)
    {
      fprintf(stderr, "divx_init: already started\n");
      return;
    }

  divx.pixmap = pixmap;

  if (debug)
    {
      fprintf(stderr, "DivX init...\n");
      switch(divx.codec)
        {
#ifdef HAVE_DIVX4LINUX
        case DIVX4LINUX:
            codec_name = "DivX";
            break;
#endif

#ifdef HAVE_FFMPEG
          case FFMPEG_MPEG4:
            codec_name = "FFMPEG - MPEG4\n";
            break;
          case FFMPEG_MPEG1:
            codec_name = "FFMPEG - MPEG1VIDEO\n";
          break;
#endif

#ifdef HAVE_XVID
        case XVID:
            codec_name = "XviD\n";
            break;
#endif
        }
      fprintf(stderr, "divx.codec             = %s\n", codec_name);
      fprintf(stderr, "divx.pixmap            = %d\n", divx.pixmap);
      fprintf(stderr, "divx.width             = %d\n", divx.width);
      fprintf(stderr, "divx.height            = %d\n", divx.height);
      fprintf(stderr, "divx.bitrate           = %ld\n", divx.bitrate);
      fprintf(stderr, "divx.quality           = %d\n", divx.quality);
      fprintf(stderr, "divx.min_quantizer     = %d\n", divx.min_quantizer);
      fprintf(stderr, "divx.max_quantizer     = %d\n", divx.max_quantizer);
      fprintf(stderr, "divx.audio_buffer_size = %d\n", divx.audio_buffer_size);
      fprintf(stderr, "divx.stereo            = %d\n", divx.stereo);
      fprintf(stderr, "divx.compress_audio    = %d\n", divx.compress_audio);
      fprintf(stderr, "divx.mp3_bitrate       = %d\n", divx.mp3_bitrate);

      fprintf(stderr, "divx.mp3_quality       = %d\n", divx.mp3_quality);
      fprintf(stderr, "divx.mp3_vbr           = %d\n", divx.mp3_vbr);
      fprintf(stderr, "divx.mp3_vbr_quality   = %d\n", divx.mp3_vbr_quality);
#ifdef SYNCHRO
      fprintf(stderr, "divx.fps               = %d\n", divx.fps);
      fprintf(stderr, "divx.maxgap            = %f\n", divx.maxgap);
#endif
      fprintf(stderr, "divx.display_frames    = %d\n", divx.display_frames);
    }

  prev_w = width;
  prev_h = height;
  width = divx.width;
  height = divx.height;

  if(width>cur_maxwidth) width = cur_maxwidth;
  if(height>cur_maxheight) height = cur_maxheight;
  if(fps / 1000.0 > (double)cur_maxfpsnum / cur_maxfpsden - 0.01) {
    fps_num = cur_maxfpsnum;
    fps_den = cur_maxfpsden;
  } else {
    fps_num = fps;
    fps_den = 1000;
  }

  if (debug)
    fprintf(stderr, "divx_init: requested video size: w = %d, h = %d\n", width, height);

  video_get_capture_size(&prev_w_capture, &prev_h_capture);
  video_set_capture_size(width, height);

  if (debug)
    fprintf(stderr, "divx_init: allocated video size: w = %d, h = %d\n", width, height);
  
  if (plugin)
    {
      height -= 2 * (int)(plugin->get_ybar() * (double)(height) / cur_maxheight + 0.5);
    }
  if (debug)
    fprintf(stderr, "divx_init: encoded video size:   w = %d, h = %d\n", width, height);

  switch(divx.codec)
    {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      // Check for DivX codec version
#if defined(ENC_OPT_VERSION) && defined (ENCORE_VERSION)
      version_binary = encore(0, ENC_OPT_VERSION, 0, 0);
      if (debug)
        fprintf(stderr, "divx_init: divx4linux DivX encoder version = %d\n", version_binary);
      if (version_binary != ENCORE_VERSION)
        {
          //...interfaces are incompatible, encoding can't be performed
          fprintf(stderr, "divx4linux DivX encoder claim its version is %d and I was expecting encoder version %d: I give up...\n",
                  version_binary, ENCORE_VERSION);
	  divx_stop();
          return;
        }
#else
      fprintf(stderr, "divx_init: Warning: at compile time, divx4linux was too old to perform version checking\n");
#endif

#if ENCORE_VERSION >= 20021024
#define FOURCC(A, B, C, D) ( ((uint8_t) (A)) | (((uint8_t) (B))<<8) | (((uint8_t) (C))<<16) | (((uint8_t) (D))<<24) )

    if ((settings = malloc(sizeof(SETTINGS)))==NULL) {
      fprintf(stderr,"out of memory");
      divx_stop();
      return;
    }
    if ((format = malloc(sizeof(DivXBitmapInfoHeader)))==NULL) {
      fprintf(stderr,"out of memory");
      divx_stop();
      return; 
    }
    memset (settings, 0, sizeof(SETTINGS));
    memset (format, 0, sizeof(DivXBitmapInfoHeader));

    settings->vbr_mode = RCMODE_502_1PASS;
    settings->bitrate = divx.bitrate;
    settings->quantizer = (float)(divx.bitrate/1000);
    settings->use_bidirect = 0;
    settings->input_clock = fps_num;
    settings->input_frame_period = fps_den;
    settings->quality = divx.quality;
    settings->complexity_modulation = 0.5;
    settings->enable_crop = 0;
    settings->enable_resize = 0;

    format->biSize = sizeof(DivXBitmapInfoHeader);
    format->biWidth = width;
    format->biHeight = height;
#else
      memset(&param, 0, sizeof(param));
      param.x_dim = width;
      param.y_dim = height;
      param.framerate = (float)fps_num/fps_den;
      param.min_quantizer = divx.min_quantizer;
      param.max_quantizer = divx.max_quantizer;
      param.quality = divx.quality; // (1 - fastest, 5 - best)
      param.bitrate = divx.bitrate; 

      ret = encore(0, ENC_OPT_INIT, &param, 0);
      if (ret != ENC_OK)
        {
          fprintf(stderr, "divx_init: divx4linux encore initialization failed\n");
	  divx_stop();
          return;
        }

      handle=param.handle;
#endif
      enc_frame = (unsigned char *)malloc(ENC_FRAME_SIZE);
      if (enc_frame == NULL)
        {
          fprintf(stderr, "divx_init: couldn't allocate memory for encoded frame: releasing divx4linux DivX encoder...\n");
          divx_stop();
          return;
        }
      memset(&frame, 0, sizeof(frame));
      frame.bitstream = enc_frame;
      
#if ENCORE_VERSION >= 20021024 
     switch(divx.pixmap)
        {
        case VIDEO_GRAY:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_GRAY\n");
          format->biCompression = FOURCC('Y','V','1','2');
          format->biBitCount = 0;
          divx.convert = gray_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB08:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB08\n");
          format->biCompression = 0;
          format->biBitCount = 32;
          divx.convert = hi240_to_rgb32;
          convert_buf = malloc(width*height*4);
          break;
        case VIDEO_RGB15:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB15\n");
          format->biCompression = FOURCC('Y','V','1','2');
          format->biBitCount = 0;
          divx.convert = rgb15_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB16:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB16\n");
          format->biCompression = FOURCC('Y','V','1','2');
          format->biBitCount = 0;
          divx.convert = rgb16_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB24:
          fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB24 *** FIX ME: Can't test this config ***\n");
          format->biCompression = 0;
          format->biBitCount = 24;
          divx.convert = NULL;
          convert_buf = NULL;
          break;
        case VIDEO_RGB32:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB32\n");
          format->biCompression = 0;
          format->biBitCount = 32;
          // Looks like there is a bug in the divx4linux encoder in RGB32 mode
          // we need to rewrite the image buffer down top....
          divx.convert = rgb32_top_down;
          convert_buf = malloc(width*height*4);
          break;
        case VIDEO_YUY2:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_YUY2\n");
          format->biCompression = FOURCC('Y','U','Y','2');
          format->biBitCount = 0;
          divx.convert = NULL;
          convert_buf = NULL;
          break;
	default:
	  fprintf(stderr,"format not supported with divx4linux\n");
	  divx_stop();
	  return;
        }


      ret = encore((void*)&handle, ENC_OPT_INIT, format, settings);
       if (ret != ENC_OK)
         {
           fprintf(stderr, "divx_init: divx4linux encore initialization failed\n");
	   fprintf(stderr, "error value is :%d\n",ret);
	   divx_stop();
           return;
         }

      if (divx.convert != NULL)
        {
          if (convert_buf == NULL)
            {
              fprintf(stderr, "divx_init: couldn't allocate memory for intermediate frame: releasing divx4linux DivX encoder...\n");
              divx_stop();
              return;
            }
        }
      break;
#else
      memset(&frame, 0, sizeof(frame));
      frame.bitstream = enc_frame;
      switch(divx.pixmap)
        {
        case VIDEO_GRAY:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_GRAY\n");
          frame.colorspace = ENC_CSP_YV12;
          divx.convert = gray_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB08:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB08\n");
          frame.colorspace = ENC_CSP_YV12;
          divx.convert = hi240_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB15:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB15\n");
          frame.colorspace = ENC_CSP_YV12;
          divx.convert = rgb15_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB16:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB16\n");
          frame.colorspace = ENC_CSP_YV12;
          divx.convert = rgb16_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB24:
          fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB24 *** FIX ME: Can't test this config ***\n");
          frame.colorspace = ENC_CSP_RGB24;
          divx.convert = NULL;
          convert_buf = NULL;
          break;
        case VIDEO_RGB32:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_RGB32\n");
#if defined(ENC_CSP_RGB32)
          frame.colorspace = ENC_CSP_RGB32;
          // Looks like there is a bug in the divx4linux encoder in RGB32 mode
          // we need to rewrite the image buffer down top....
          divx.convert = rgb32_top_down;
          convert_buf = malloc(width*height*4);
#else
          frame.colorspace = ENC_CSP_YV12;
          divx.convert = rgb32_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
#endif
          break;
        case VIDEO_YUY2:
          if (debug)
            fprintf(stderr, "divx4linux: pixmap = VIDEO_YUY2\n");
          frame.colorspace = ENC_CSP_YUY2;
          divx.convert = NULL;
          convert_buf = NULL;
          break;
	default:
	  fprintf(stderr,"format not supported with divx4linux");
	  divx_stop();
	  return;
        }

      if (divx.convert != NULL)
        {
          if (convert_buf == NULL)
            {
              fprintf(stderr, "divx_init: couldn't allocate memory for intermediate frame: releasing divx4linux DivX encoder...\n");
              divx_stop();
              return;
            }
        }
      frame.mvs = 0;
      break;
#endif
#endif // HAVE_DIVX4LINUX

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
    case FFMPEG_MPEG1:
      avcodec_init();
      fprintf(stderr, "divx_init: initializing ffmpeg libavcodec version %d (build %d)\n", avcodec_version(), avcodec_build());
      if (divx.codec == FFMPEG_MPEG4)
        {
          register_avcodec(&mpeg4_encoder);
          codec = avcodec_find_encoder(CODEC_ID_MPEG4);
        }
      else if (divx.codec == FFMPEG_MPEG1)
        {
          register_avcodec(&mpeg1video_encoder);
          codec = avcodec_find_encoder(CODEC_ID_MPEG1VIDEO);
        }

      if (codec == NULL)
         {
            fprintf(stderr, "divx_init: couldn't find %s codec\n", codec_name);
	    divx_stop();
            return;
          }

      if (codec_context == NULL)
        codec_context = handle = avcodec_alloc_context();
      picture_420p = avcodec_alloc_frame();

      /* All default parameters are borrowed from MPlayer */
      if (divx.codec == FFMPEG_MPEG4)
        codec_context->codec_id            = CODEC_ID_MPEG4;
      else if (divx.codec == FFMPEG_MPEG1)
        codec_context->codec_id            = CODEC_ID_MPEG1VIDEO;
      codec_context->codec_type            = CODEC_TYPE_VIDEO;
      codec_context->bit_rate              = divx.bitrate;
      codec_context->bit_rate_tolerance    = divx.bitrate*10;
      codec_context->width                 = width;
      codec_context->height                = height;
#if LIBAVCODEC_BUILD >= 4662
      codec_context->frame_rate		   = fps_num;
      codec_context->frame_rate_base       = fps_den;
#else
      codec_context->frame_rate            = fps_num*FRAME_RATE_BASE/fps_den;
#endif
      codec_context->qmin                  = divx.min_quantizer;
      codec_context->qmax                  = divx.max_quantizer;
      codec_context->max_qdiff             = 3;
      codec_context->qcompress             = 0.5;
      codec_context->qblur                 = 0.5;
      codec_context->max_b_frames          = 0;
      codec_context->b_quant_factor        = 1.25;
      codec_context->rc_strategy           = 2;
      codec_context->b_frame_strategy      = 0;
#ifdef CODEC_FLAG_PART
      codec_context->b_quant_offset        = 1.25;
      codec_context->luma_elim_threshold   = 0;
      codec_context->chroma_elim_threshold = 0;
      codec_context->rtp_payload_size      = 0;
      /*
      if(param_packet_size )
        codec_context->rtp_mode            = 1;
      */
      codec_context->rtp_mode              = 0;
      codec_context->strict_std_compliance = 0;
#endif
      codec_context->i_quant_factor        = 0.8;
      codec_context->i_quant_offset        = 0.0;
      codec_context->rc_qsquish            = 1.0;
      codec_context->rc_qmod_amp           = 0;
      codec_context->rc_qmod_freq          = 0;
      codec_context->rc_eq                 = "tex^qComp";
      codec_context->rc_max_rate           = 0*1000;
      codec_context->rc_min_rate           = 0*1000;
      codec_context->rc_buffer_size        = 0*1000;
      codec_context->rc_buffer_aggressivity= 1.0;
      codec_context->rc_initial_cplx       = 0;

      /*
      p= lavc_param_rc_override_string;
      for(i=0; p; i++){
        int start, end, q;
        int e=sscanf(p, "%d,%d,%d", &start, &end, &q);
        if(e!=3){
          mp_msg(MSGT_MENCODER,MSGL_ERR,"error parsing vrc_q\n");
          return 0;
        }
        codec_context->rc_override=
          realloc(codec_context->rc_override, sizeof(RcOverride)*(i+1));
        codec_context->rc_override[i].start_frame= start;
        codec_context->rc_override[i].end_frame  = end;
        if(q>0){
          codec_context->rc_override[i].qscale= q;
          codec_context->rc_override[i].quality_factor= 1.0;
        }
        else{
          codec_context->rc_override[i].qscale= 0;
          codec_context->rc_override[i].quality_factor= -q/100.0;
        }
        p= strchr(p, '/');
        if(p) p++;
      }
      codec_context->rc_override_count=i;
      */


      codec_context->mpeg_quant            = 0;
      codec_context->dct_algo              = 0;
      codec_context->idct_algo             = 0;
      codec_context->lumi_masking          = 0.0;
      codec_context->temporal_cplx_masking = 0.0;
      codec_context->spatial_cplx_masking  = 0.0;
      codec_context->p_masking             = 0.0;
      codec_context->dark_masking          = 0.0;

      /*
      if (lavc_param_aspect != 0.0)
        {
          // 625 means CIF
          if (lavc_param_aspect == (float)(4.0/3.0))
            codec_context->aspect_ratio_info = FF_ASPECT_4_3_625;
          else if (lavc_param_aspect == (float)(16.0/9.0))
            codec_context->aspect_ratio_info = FF_ASPECT_16_9_625;
          else if (lavc_param_aspect == (float)(221.0/100.0))
            {
              codec_context->aspect_ratio_info = FF_ASPECT_EXTENDED;
              codec_context->aspected_width = 221;
              codec_context->aspected_height = 100;
            }
          else
            {
              printf("Unsupported aspect ratio (%f)\n", lavc_param_aspect);
            }
          mp_dbg(MSGT_MENCODER, MSGL_DBG2, "aspect_ratio_info: %d\n", codec_context->aspect_ratio_info);
          mp_dbg(MSGT_MENCODER, MSGL_DBG2, "par_width: %d\n", codec_context->aspected_width);
          mp_dbg(MSGT_MENCODER, MSGL_DBG2, "par_height: %d\n", codec_context->aspected_height);
        }
      */
      codec_context->gop_size           = 250; // lavc_param_keyint
      codec_context->flags              = 0;   // MPlayer default

      /* 4mv is currently buggy with B frames */
      /*
      if (lavc_param_vmax_b_frames > 0 && lavc_param_v4mv) {
        printf("4MV with B-Frames not yet supported -> 4MV disabled\n");
        lavc_param_v4mv = 0;
      }
      codec_context->flags|= lavc_param_v4mv ? CODEC_FLAG_4MV : 0;
#ifdef CODEC_FLAG_PART
      codec_context->flags|= lavc_param_data_partitioning;
#endif
      if(lavc_param_gray) codec_context->flags|= CODEC_FLAG_GRAY;
      if(lavc_param_normalize_aqp) codec_context->flags|= CODEC_FLAG_NORMALIZE_AQP;
      if(lavc_param_interlaced_dct) codec_context->flags|= CODEC_FLAG_INTERLACED_DCT;
      */
      codec_context->me_method             = Motion_Est_QTab[divx.quality];

      /* fixed qscale :p */
      /*
      if (lavc_param_vqscale)
        {
          //printf("Using constant qscale = %d (VBR)\n", lavc_param_vqscale);
          lavc_venc_context->flags        |= CODEC_FLAG_QSCALE;
          lavc_venc_context->quality       = 0;
        }
      */

      // We are using only a one pass compression
      codec_context->flags |= CODEC_FLAG_PASS1;

      switch(divx.pixmap)
        {
        case VIDEO_GRAY:
          if (debug)
	    fprintf(stderr, "ffmpeg: pixmap = VIDEO_GRAY\n");
          divx.convert = gray_to_yuv420p;
          break;
        case VIDEO_RGB08:
          if (debug)
	    fprintf(stderr, "ffmpeg: pixmap = VIDEO_RGB08\n");
          divx.convert = hi240_to_yuv420p;
          break;
        case VIDEO_RGB15:
          if (debug)
            fprintf(stderr, "ffmpeg: pixmap = VIDEO_RGB15\n");
          divx.convert = rgb15_to_yuv420p;
          break;
        case VIDEO_RGB16:
          if (debug)
            fprintf(stderr, "ffmpeg: pixmap = VIDEO_RGB16\n");
          divx.convert = rgb16_to_yuv420p;
          break;
        case VIDEO_RGB24:
          fprintf(stderr, "ffmpeg: pixmap = VIDEO_RGB24  *** FIX ME: Can't test this config ***\n");
          divx.convert = rgb24_to_yuv420p;
          break;
        case VIDEO_RGB32:
          if (debug)
            fprintf(stderr, "ffmpeg: pixmap = VIDEO_RGB32\n");
          divx.convert = rgb32_to_yuv420p;
          break;
        case VIDEO_YUY2:
          if (debug)
            fprintf(stderr, "ffmpeg: pixmap = VIDEO_YUY2\n");
          divx.convert = yuv422_to_yuv420p;
          break;
	default:
	  fprintf(stderr,"format not supported with XviD\n");
	  divx_stop();
	  return;
        }

      // Allocate ffmpeg working buffer
      convert_buf = malloc((width*height*3)/2); /* size for planar YUV 420 */
      if (convert_buf == NULL)
        {
          fprintf(stderr, "divx_init: couldn't allocate memory for ffmpeg working buffer: releasing ffmpeg DivX encoder...\n");
              divx_stop();
            }

      // Fill in picture fields
      picture_420p->data[0] = convert_buf;
      picture_420p->data[1] = picture_420p->data[0] + (codec_context->width*codec_context->height);
      picture_420p->data[2] = picture_420p->data[1] + (codec_context->width*codec_context->height) / 4;
      picture_420p->linesize[0] = codec_context->width;
      picture_420p->linesize[1] = codec_context->width / 2;
      picture_420p->linesize[2] = codec_context->width / 2;

      if (avcodec_open(codec_context, codec) < 0)
        {
          fprintf(stderr, "divx_init: couldn't open %s codec\n", codec_name);
	  divx_stop();
          return;
        }

      if (codec_context->codec->encode == NULL)
        {
          fprintf(stderr, "divx_init: %s codec initialization failed (ctx->codec->encode == NULL)!\n", codec_name);
	  divx_stop();
          return;
        }

      enc_frame = (unsigned char *)malloc(ENC_FRAME_SIZE);
      if (enc_frame == NULL)
        {
          fprintf(stderr, "divx_init: couldn't allocate memory for encoded frame: releasing ffmpeg DivX encoder...\n");
          divx_stop();
          return;
        }

      break;
#endif // HAVE_FFMPEG

#ifdef HAVE_XVID
    case XVID:
      memset(&xinit, 0, sizeof(xinit));
      xvid_init(NULL, 0, &xinit, NULL);
      fprintf(stderr, "divx_init: initializing XviD core version %d\n", xinit.core_build);
      if (debug)
        fprintf(stderr, "divx_init: XviD API version = %d.%d\n", xinit.api_version>>16, xinit.api_version&0xFF);
      if (xinit.api_version != API_VERSION)
        {
          //...interfaces are incompatible, encoding can't be performed
          fprintf(stderr, "XviD core claim its API version is %d.%d and I was expecting API version %d%.d: I give up...\n",
                  xinit.api_version>>16, xinit.api_version&0xFF,
                  API_VERSION>>16, API_VERSION&0xFF);
	  divx_stop();
          return;
        }

      memset(&xparam, 0, sizeof(xparam));
      xparam.width = width;
      xparam.height = height;
      xparam.fincr = FRAMERATE_INCR;
      xparam.fbase = fps_num*FRAMERATE_INCR/fps_den;
      xparam.rc_bitrate = divx.bitrate;
      xparam.rc_reaction_delay_factor = -1; // forces default value
      xparam.rc_averaging_period = -1; // forces default value
      xparam.rc_buffer = -1; // forces default value
      xparam.max_quantizer = divx.max_quantizer;
      xparam.min_quantizer = divx.min_quantizer;
      xparam.max_key_interval = 10*fps_num/fps_den; // default value is too big
      ret = xvid_encore(NULL, XVID_ENC_CREATE, &xparam, NULL);
      if (ret != XVID_ERR_OK)
        {
          fprintf(stderr, "divx_init: XviD encore initialization failed\n");
	  divx_stop();
          return;
        }
      handle = xparam.handle;

      memset(&xframe, 0, sizeof(xframe));
      xframe.general = xgeneral_presets[divx.quality];
      xframe.motion = xmotion_presets[divx.quality];
      xframe.quant = 0;
      xframe.intra = -1;

      divx.convert = NULL;
      convert_buf = NULL;
      
      switch(divx.pixmap)
        {
        case VIDEO_GRAY:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_RGB15\n");
	  xframe.colorspace = XVID_CSP_YV12;
          divx.convert = gray_to_yuv420p;
          convert_buf = malloc((width*height*3)/2);
          break;
        case VIDEO_RGB08:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_RGB08\n");
	  xframe.colorspace = XVID_CSP_RGB32;
          divx.convert = hi240_to_rgb32;
          convert_buf = malloc(width*height*4);
          break;
        case VIDEO_RGB15:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_RGB15\n");
          xframe.colorspace = XVID_CSP_RGB555;
          break;
        case VIDEO_RGB16:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_RGB16\n");
          xframe.colorspace = XVID_CSP_RGB565;
          break;
        case VIDEO_RGB24:
          fprintf(stderr, "XviD: pixmap = VIDEO_RGB24 *** FIX ME: Can't test this config ***\n");
          xframe.colorspace = XVID_CSP_RGB24;
          break;
        case VIDEO_RGB32:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_RGB32\n");
#if 0
	  // does not work with xvid 0.9.1
          xframe.colorspace = XVID_CSP_RGB32 | XVID_CSP_VFLIP;
#else
          xframe.colorspace = XVID_CSP_RGB32;
	  divx.convert = rgb32_top_down;
	  convert_buf = malloc(width*height*4);
#endif
          break;
        case VIDEO_YUY2:
          if (debug)
            fprintf(stderr, "XviD: pixmap = VIDEO_YUY2\n");
          xframe.colorspace = XVID_CSP_YUY2;
          break;
	default:
	  fprintf(stderr,"format not supported with XviD\n");
	  divx_stop();
	  return;
        }

      enc_frame = (unsigned char *)malloc(ENC_FRAME_SIZE);
      if (enc_frame == NULL)
        {
          fprintf(stderr, "divx_init: couldn't allocate memory for encoded frame: releasing XviD encoder...\n");
          divx_stop();
          return;
        }
      xframe.bitstream = enc_frame;
      break;
#endif // HAVE_XVID
    }

  // Open and setup AVI file
  filename = malloc(256);
  get_divxfilepath(&divx_path);
  strcpy(filename, divx_path);
  strcat(filename, "/");
  strcat(filename, divx_file);
  if(debug){
      fprintf(stderr, "divx_file    = %s\n",divx_file);
      fprintf(stderr, "divx_path    = %s\n",divx_path);
      fprintf(stderr, "filename     = %s\n",filename);
  }
  avifile = AVI_open_output_file(filename);
  if(divx.sub) {
    strcpy(filename+strlen(filename)-3,"sub");
    subfile=fopen(filename,"w");
    if(subfile==NULL) perror("fopen sub");
    lastsubend=0;
  }
  if(subfile) {
    fprintf(subfile,"FORMAT=%f\n\n",(double)fps_num/fps_den);
  }
  free(filename);

  if (avifile == NULL)
    {
      AVI_print_error("divx_init: AVI_open_output_file:");
      divx_stop();
      return;
    }

  switch (divx.codec)
  {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
#if (ENCORE_MAJOR_VERSION >= 5010)
      fourcc = "DX50";
#else
      fourcc = "DIVX";
#endif
      break;
#endif

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
      fourcc = "DIVX";
      break;
    case FFMPEG_MPEG1:
      fourcc = "MPEG";
      break;
#endif

#ifdef HAVE_XVID
    case XVID:
      fourcc = "XVID";
      break;
#endif
  }
  AVI_set_video(avifile, width, height, fps_num, fps_den, fourcc);
  
#ifdef NOAUDIO
  AVI_set_audio(avifile, 0 /* channels*/,
                         0 /* rate */,
                         0 /* bits */,
                         0 /* formats */,
                         0 /* brate */,
                         0 /* framesize */,
                         0 /* vbr */);
#else

 /* the stereo mode could be automatically detected by the card !*/
 if(divx.stereo) achans = 2; else achans = 1;

#ifdef HAVE_LAME 
 if (divx.compress_audio)
   {
     gfp = lame_init();
  
     lame_set_num_channels(gfp, achans);
  
     lame_set_in_samplerate(gfp, DEC_FREQ);
     lame_set_compression_ratio(gfp, 0.0); 
     if(!divx.mp3_vbr)
       lame_set_brate(gfp, divx.mp3_bitrate); 
     else {
       lame_set_VBR(gfp, vbr_mtrh);
       lame_set_VBR_q(gfp,divx.mp3_vbr_quality);
     }

     // mode = 0,1,2,3 = stereo, jstereo, dual channel (not supported), mono
     // default: lame picks based on compression ration and input channels  
     if (achans == 1)
       lame_set_mode(gfp, 3);
     else
       lame_set_mode(gfp, 1);
     
     // quality=0..9.  0=best (very slow).  9=worst.  
     // recommended:  2     near-best quality, not too slow
     //               5     good quality, fast
     //               7     ok quality, really fast  
     lame_set_quality(gfp, divx.mp3_quality);

     if (lame_init_params(gfp) < 0)
       {
         fprintf(stderr, "divx_init: lame_init_params failed\n");
         divx_stop();
         return;
       }

     lame_buffer_size = (1.25*divx.audio_buffer_size)+7200;
     lame_buffer = malloc(lame_buffer_size);
     if (lame_buffer == NULL)
       {
         fprintf(stderr, "divx_init: could not allocate lame buffer\n");
         divx_stop();
         return;
       }

     aformat = WAVE_FORMAT_MPEGLAYER3;
     AVI_set_audio(avifile, achans, DEC_FREQ, 16, aformat, (divx.mp3_bitrate * 1000)/8,lame_get_framesize(gfp),divx.mp3_vbr);
     if (debug)
       fprintf (stderr, "divx_init: audio: WAVE_FORMAT_MPEGLAYER3\n");
   }
 else
   {
     aformat = WAVE_FORMAT_PCM;
     AVI_set_audio(avifile, achans, DEC_FREQ, 16, aformat, 0, 0, 0);
     if (debug)
       fprintf (stderr, "divx_init: audio: WAVE_FORMAT_PCM\n");
   }
#else
 aformat = WAVE_FORMAT_PCM;
 AVI_set_audio(avifile, achans, DEC_FREQ, 16, aformat, 0, 0, 0);
 if (debug)
   fprintf (stderr, "divx_init: audio: WAVE_FORMAT_PCM\n");
#endif // HAVE_LAME

#endif // NOAUDIO

 // Let's go
  video_frames = 0;
  slw = 0;
#ifndef NOAUDIO  
  audio_frames = 0;
  audio_buffers = 0;
#ifdef SYNCHRO
  nbsamplesaudios = 0;
  dropped_video_frames = 0;
  dropped_audio_samples = 0;
  dropped_audio_buffers = 0;
#endif
  gettimeofday(&tv, NULL);
  if(audiostamp==0.0) audiostamp=tv.tv_usec/1000000.0+tv.tv_sec;
  if(videostampmin==0.0)
    videostampmin=videostampmax=tv.tv_usec/1000000.0+tv.tv_sec;
  if (!(plugin && plugin->is_nagraudio_used()))
    divx_audio_start();
#endif

  if (plugin)
    plugin->record_movie(1);

  time(&t);
  printf("Recording started %s", ctime(&t));
  gettimeofday(&tv_start, &tz);

  divx_in_progress = 1;
}

void
divx_encode(unsigned char *src_frame)
{
#ifdef HAVE_DIVX4LINUX
   ENC_RESULT res;
#endif
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_XVID)
  int ret;
#endif
#ifdef HAVE_FFMPEG
  int enc_size;
#endif
  int flag;
  unsigned char *y, *u, *v;


  /* only the video thread can run divx_stop */
  if(slw) 
    {
      fprintf(stderr,"last AVI_write_status = %d\n",slw);
      AVI_print_error("reason");
      divx_stop();
      displaynextframe = 0;
      if(AVI_errno==AVI_ERR_SIZELIM) {
	char *filename;
	filename = snap_filename
	  ("movie",(channels) ? channels[cur_sender]->name:"unknown",
	   "avi");
	return divx_init(filename, divx.pixmap, prev_w, prev_h);
      }
      return;
    }
  
#ifdef SYNCHRO
  if(((double)video_frames * fps_den / fps_num - videostampmax
      + divx.delay > 
      (double)nbsamplesaudios / DEC_FREQ - audiostamp + divx.maxgap)
     || (videostampmin > audiostamp + 4.0) ) { //if audio is blocked
    dropped_video_frames++;
    return;
  }
#endif

  switch (divx.codec)
    {
#ifdef HAVE_DIVX4LINUX
 #if ENCORE_VERSION >= 20021024
    case DIVX4LINUX:
      if (divx.convert != NULL)
        {
          int size = format->biWidth * format->biHeight;
          y = convert_buf;
          u = convert_buf+(5*size)/4;
          v = convert_buf+size;
          divx.convert(y, u, v, src_frame, format->biWidth, format->biHeight);
          frame.image = convert_buf;
        }
      else
        frame.image = src_frame;

      ret = encore(handle, ENC_OPT_ENCODE, &frame, &res);
      if (ret != ENC_OK)
        {
          fprintf(stderr, "divx_encode: error encoding frame (divx4linux)!\n");
          slw = AVI_write_frame(avifile, frame.bitstream, 0, AVI_FLAG_NONE);
        }
      else
        {
          flag = (res.cType == 'I') ? AVI_FLAG_KEYFRAME : AVI_FLAG_NONE;
//	fprintf(stderr, "frame size : %d\n",res.iTotalBits);
	 // flag = AVI_FLAG_NONE;
          slw = AVI_write_frame(avifile, frame.bitstream, frame.length, flag);
        }
      break;

 #else
    case DIVX4LINUX:
      if (divx.convert != NULL)
        {
          int size = param.x_dim * param.y_dim;
          y = convert_buf;
          u = convert_buf+(5*size)/4;
          v = convert_buf+size;
          divx.convert(y, u, v, src_frame, param.x_dim, param.y_dim);
          frame.image = convert_buf;
        }
      else
        frame.image = src_frame;

      ret = encore(handle, ENC_OPT_ENCODE, &frame, &res);
      if (ret != ENC_OK)
        {
          fprintf(stderr, "divx_encode: error encoding frame (divx4linux)!\n");
          slw = AVI_write_frame(avifile, frame.bitstream, 0, AVI_FLAG_NONE);
        }
      else
        {
          flag = res.is_key_frame ? AVI_FLAG_KEYFRAME : AVI_FLAG_NONE;
          slw = AVI_write_frame(avifile, frame.bitstream, frame.length, flag);
        }
      break;
 #endif
#endif // HAVE_DIVX4LINUX

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
	case FFMPEG_MPEG1:
      if (divx.convert != NULL)
        {
          y = picture_420p->data[0];
          u = picture_420p->data[1];
          v = picture_420p->data[2];
        divx.convert(y, u, v, src_frame, codec_context->width, codec_context->height);
        }

      picture_420p->key_frame = -1;
      enc_size = avcodec_encode_video(codec_context, enc_frame, ENC_FRAME_SIZE, picture_420p);
      if (enc_size > 0)
        {
          flag = codec_context->coded_frame->key_frame ? AVI_FLAG_KEYFRAME : AVI_FLAG_NONE;
          slw = AVI_write_frame(avifile, enc_frame, enc_size, flag);
        }
      else
       {
          fprintf(stderr, "divx_encode: error encoding frame (ffmpeg)!\n");
          slw = AVI_write_frame(avifile, enc_frame, 0, AVI_FLAG_NONE);
       }
      break;
#endif // HAVE_FFMPEG

#ifdef HAVE_XVID
    case XVID:
      if(divx.convert) {
	int size = xparam.width * xparam.height;
	y = convert_buf;
	u = convert_buf+(5*size)/4;
	v = convert_buf+size;
	divx.convert(y, u, v, src_frame, xparam.width, xparam.height);
	xframe.image = convert_buf;
      }
      else {
	xframe.image = src_frame;
      }
      xframe.intra = -1;
      ret = xvid_encore(handle, XVID_ENC_ENCODE, &xframe, NULL);
      if (ret != XVID_ERR_OK)
        {
          fprintf(stderr, "divx_encode: error encoding frame (XviD)!\n");
          slw = AVI_write_frame(avifile, xframe.bitstream, 0, AVI_FLAG_NONE);
        }
      else
        {
          flag = (xframe.intra > 0) ? AVI_FLAG_KEYFRAME : AVI_FLAG_NONE;
          slw = AVI_write_frame(avifile, xframe.bitstream, xframe.length, flag);
        }
      break;
#endif // HAVE_XVID
    }
  video_frames++;

}

void
divx_stop(void)
{
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_XVID)
  int ret;
#endif
  double elapsed_time;
  struct timeval tv_stop;
  struct timezone tz;
  time_t t;

  divx_in_progress = 0;

  gettimeofday(&tv_stop, &tz);
  time(&t);
  printf("Recording stoped %s", ctime(&t));

  if (debug)
    fprintf(stderr, "divx_stop...\n");

#ifndef NOAUDIO
  divx_audio_stop();
  if (debug)
    fprintf(stderr, "divx_stop: audio stoped\n");

  if (plugin && plugin->is_nagraudio_used())
    plugin->wait_write_audio();

#ifdef HAVE_LAME
  if (divx.compress_audio)
    lame_stop();
#endif // HAVE_LAME
#endif

  // Set the real FPS
  elapsed_time = ((double)tv_stop.tv_sec + (double)tv_stop.tv_usec*1.0e-6)
    - ((double)tv_start.tv_sec + (double)tv_start.tv_usec*1.0e-6);

#ifndef SYNCHRO
  fps_den = 1000000;
  fps_num = (double)video_frames * fps_den / elapsed_time + 0.5;

  if (avifile) {
    avifile->fps_num = fps_num;
    avifile->fps_den = fps_den;
  }
#else
  if(avifile && (avifile->fps_num != fps_num || avifile->fps_den != fps_den))
    fprintf(stderr,"ATTENTION fps has changed during recording\n");
#endif

#ifndef NOAUDIO
#ifndef SYNCHRO
  printf("%ld frames recorded in %f seconds => FPS = %f\n%ld audio frames recorded (%ld audio buffers)\n", video_frames, elapsed_time, (double)fps_num/fps_den, audio_frames,audio_buffers);
#else
  printf("recording time = %f seconds,  ", elapsed_time);
  printf("fps = %f\n",(double)fps_num/fps_den);
  printf("%ld video frames recorded (%f seconds)\n", video_frames,
	 (double)video_frames*fps_den/fps_num);
  printf("%ld audio frames recorded (%ld audio buffers, %f seconds)\n", audio_frames,audio_buffers,(double)nbsamplesaudios / DEC_FREQ);
  printf("%ld dropped video frames (fpscod=%f fpscapt=%f)\n",
	 dropped_video_frames, (double)video_frames / elapsed_time,
	 ((double)video_frames + dropped_video_frames) / elapsed_time);
  printf("%ld dropped audio buffers (%f seconds)\n",
	 dropped_audio_buffers, (double)dropped_audio_samples / DEC_FREQ);
  if(video_frames==0) {
    printf("@@@@ No video frames recorded\n");
    printf("@@@@ Recording must be done in grabdisplay mode !!!\n");
  }
  else if(dropped_audio_buffers*20>audio_frames) {
    printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
    printf("@@ You computer is too slow to do such recording @@\n");
    printf("@@ try to decrease image dimension               @@\n");
    printf("@@ or the framerate (to %.1ffps ?)               @@\n",
	   (double)video_frames / elapsed_time * 1.05);
    printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
  }
  else if(dropped_video_frames*20>video_frames) {
    printf("You could try record to the better fps=%.1f\n",
	   ((double)video_frames + dropped_video_frames)
	   / elapsed_time * 0.95);
  }
#endif
#else
  printf("%ld frames recorded in %f seconds => FPS = %f\n", video_frames, elapsed_time,(double)fps_num/fps_den);
#endif
  
  
  // Release allocated buffers
  switch(divx.codec)
    {
#ifdef HAVE_DIVX4LINUX
    case DIVX4LINUX:
      if (handle != NULL)
        {
          /* Release the encoder */
          ret = encore(handle, ENC_OPT_RELEASE, 0, 0);
          if (ret != ENC_OK)
            {
              fprintf(stderr, "divx_stop: divx4linux codec release failed\n");
            }
          else if (debug)
            fprintf(stderr, "divx_stop: divx4linux codec released\n");

          handle = NULL;
        }

      if (enc_frame != NULL)
        {
          free(enc_frame);
          enc_frame = NULL;
          if (debug)
            fprintf(stderr, "divx_stop: divx4linux encoded frame freed\n");
        }
      break;
#endif // HAVE_DIVX4LINUX

#ifdef HAVE_FFMPEG
    case FFMPEG_MPEG4:
	case FFMPEG_MPEG1:
      if (codec_context != NULL)
        {
          avcodec_close(codec_context);
          free(codec_context);
          codec_context = NULL;
	  handle = NULL;
          free(picture_420p);
          picture_420p = NULL;
          if (debug)
            fprintf(stderr, "divx_stop: ffmpeg codec released\n");
        }

      if (enc_frame != NULL)
        {
          free(enc_frame);
          enc_frame = NULL;
          if (debug)
            fprintf(stderr, "divx_stop: ffmpeg encoded frame freed\n");
        }
      break;
#endif // HAVE_FFMPEG

#ifdef HAVE_XVID
    case XVID:
      if (handle != NULL)
        {
          /* Release the encoder */
          ret = xvid_encore(handle, XVID_DEC_DESTROY, NULL, NULL);
          if (ret != XVID_ERR_OK)
            {
              fprintf(stderr, "divx_stop: XviD codec release failed\n");
            }
          else if (debug)
            fprintf(stderr, "divx_stop: XviD codec released\n");

          handle = NULL;
        }

      if (enc_frame != NULL)
        {
          free(enc_frame);
          enc_frame = NULL;
          if (debug)
            fprintf(stderr, "divx_stop: XviD encoded frame freed\n");
        }
      break;
#endif // HAVE_XVID
    }

  if (convert_buf != NULL)
    {
      free(convert_buf);
      convert_buf = NULL;
      if (debug)
        fprintf(stderr, "divx_stop: convert buffer freed\n");
    }

  if (avifile)
    {
      printf("file size = %.1f Mb,  ",(unsigned)avifile->pos/1024.0/1024.0);
      printf("mean bitrate = %.3f kbps\n",(long long)(unsigned)avifile->pos*8.0/elapsed_time/1000);
      if ( AVI_close(avifile) != 0)
        AVI_print_error("divx_stop: AVI_close:");
      
      avifile = NULL;
      if (debug)
        fprintf(stderr, "divx_stop: divx file closed\n");
    }

  if(subfile)
    {
      write_sub(NULL);
      fclose(subfile);
      subfile=NULL;
    }

  if (plugin)
    plugin->record_movie(0);

  video_set_capture_size(prev_w_capture, prev_h_capture);

  if (debug)
    fprintf(stderr, "...Done\n");
}

int
is_divx_in_progress(void)
{
  return divx_in_progress;
}

int
divx_display_frames(void)
{
  //when divx_stop is started from divx_encode
  //the frame size changes, assigning xvimage to NULL
  if(!displaynextframe) { displaynextframe=1; return 0; }
  return divx.display_frames;
}

#ifndef NOAUDIO
/*****************************************************************************
 *                            RAW AUDIO FUNCTIONS
 ****************************************************************************/
void
divx_audio_start(void)
{
  int aux, caps;
  int arg = 0;
  int ret;

  audiofd = open (device, O_RDONLY);
  if (audiofd < 0)
    {
      fprintf (stderr, "audio_init: Error opening audio device\n");
      return;
    }

  if (ioctl (audiofd, SNDCTL_DSP_GETCAPS, &caps) == -1)
    {
      fprintf (stderr, "Sorry but your sound driver is too old\n");
      return;
    }

  aux = FRAGMENTS;
  if (ioctl (audiofd, SNDCTL_DSP_SETFRAGMENT, &aux) == -1)
    {
      fprintf (stderr, "Error setting audio fragments\n");
      return;
    }

  //aux = FORMAT8;
  aux = FORMAT16;
  if (ioctl (audiofd, SNDCTL_DSP_SETFMT, &aux) == -1)
    {
      fprintf (stderr, "Error setting audio format (bits)\n");
      return;
    }

  aux = achans != 1;
  if (ioctl (audiofd, SNDCTL_DSP_STEREO, &aux) == -1)
    {
      fprintf (stderr, "Error setting audio format (stereo)\n");
      return;
    }

  aux = DEC_FREQ;
  if (ioctl (audiofd, SNDCTL_DSP_SPEED, &aux) == -1)
    {
      fprintf (stderr, "Error setting audio format (frequency)\n");
      return;
    }

  ret = pthread_create (&audio_thread, NULL, audio_record, (void *) arg);
  if (ret != 0)
    fprintf (stderr, "Audio record thread failed to start.\n");
  else if (debug)
    fprintf (stderr, "Audio record thread started.\n");

}

inline void
write_audio_to_avi(short int *buffer, int size)
{
#ifdef HAVE_LAME
  int nsamples;
#endif // HAVE_LAME

  if(slw) return;
#ifdef SYNCHRO
  if(((double)nbsamplesaudios / DEC_FREQ - audiostamp > 
      (double)video_frames * fps_den / fps_num
      - videostampmin + divx.delay + divx.maxgap)
     || (audiostamp > videostampmax + 4.0)) { //if video is blocked
    dropped_audio_samples += size / achans;
    dropped_audio_buffers++;
    return;
  }
  nbsamplesaudios += size / achans;
#endif

#ifdef HAVE_LAME
  if (divx.compress_audio)
    {
      if(!divx.mp3_vbr) {
	nsamples = lame_encode(buffer, size / achans);
	slw = AVI_write_audio(avifile, lame_buffer, nsamples);
	audio_frames++;
      } else {
	/* in this case an AVI frame must contain only one MP3 frame */
	int i;
	for(i=0;i<size/achans;i+=avifile->a_framesize) {
	  if(i+avifile->a_framesize<size/achans) {
	    nsamples = lame_encode(buffer+i,avifile->a_framesize);
	    slw = AVI_write_audio(avifile, lame_buffer, nsamples);
	    if(slw) return;
	    audio_frames++;
	  } else {
	    nsamples = lame_encode(buffer+i, size / achans - i);
	    if(nsamples) {
	      slw = AVI_write_audio(avifile, lame_buffer, nsamples);
	      audio_frames++;
	    }
	  }
	}
      }
    }
  else
#endif // HAVE_LAME
    {
      slw = 
	AVI_write_audio(avifile, (char *)buffer, size * sizeof (short int));
      audio_frames++;
    }
    
    audio_buffers++;
}

static void*
audio_record(void *arg)
{
  int error;
  short int *buffer;
  struct timeval tv;

  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
  buffer = (short int *)malloc (/*BUFFER_SIZE*/divx.audio_buffer_size * sizeof (short int));
  if (buffer == NULL)
    {
      error = 1;
      goto error;
    }

  while (1)
    {
      if (read (audiofd, buffer, /*BUFFER_SIZE*/divx.audio_buffer_size * sizeof (short int)) == -1)
        {
          error = 2;
          goto error;
        }
      gettimeofday (&tv, NULL);
      audiostamp=tv.tv_usec/1000000.0+tv.tv_sec;
      write_audio_to_avi(buffer, divx.audio_buffer_size);

      pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
      pthread_testcancel ();
      pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
    }
  return NULL;

 error:
  switch (error)
    {
    case 1:
      fprintf(stderr, "audio_record: Error allocating audio buffer: exiting audio record thread.\n");
      break;
    case 2:
      fprintf(stderr, "audio_record: Error reading device: exiting audio record thread.\n");
      break;
    }
  pthread_exit ((void *)&error);
}

void
divx_audio_stop(void)
{
  void *ret;

  if (audiofd == -1)
    return;

  pthread_cancel (audio_thread);
  pthread_join (audio_thread, &ret);
  
  if (audiofd != -1)
    {
      close (audiofd);
      audiofd = -1;
    }
}


/*****************************************************************************
 *                           MP3 AUDIO FUNCTIONS
 ****************************************************************************/
#ifdef HAVE_LAME

inline static int
lame_encode(short int *raw_buffer, int samples_per_channel)
{
  int mp3_samples;

  if(achans == 1)
    mp3_samples = lame_encode_buffer(gfp, raw_buffer, raw_buffer, samples_per_channel, lame_buffer, lame_buffer_size);
  else
    mp3_samples = lame_encode_buffer_interleaved(gfp, raw_buffer, samples_per_channel, lame_buffer, lame_buffer_size);
  return  mp3_samples;
}

static void
lame_stop(void)
{
  int i;

  if(gfp == NULL) return;

  i = lame_encode_flush(gfp, lame_buffer, lame_buffer_size);

  if (avifile && !slw)
    slw = AVI_write_audio(avifile, lame_buffer, i);

  lame_close(gfp);
  gfp = NULL;

  if (lame_buffer)
    {
      free(lame_buffer);
      lame_buffer = NULL;
      lame_buffer_size = 0;
    }
}
#endif // HAVE_LAME

#else //NOAUDIO

inline void write_audio_to_avi(short int *buffer, int size){}

#endif //NOAUDIO


/*****************************************************************************
 *                SOME COLOR SPACE (AND OTHERS...) CONVERSIONS
 ****************************************************************************/
static void
yuv422_to_yuv420p(unsigned char *d, unsigned char *foo1, unsigned char *foo2, unsigned char *s, int w, int h)
{
  int  a,b;
  unsigned char *y,*u,*v;

  y = d;
  u = y + w * h;
  v = u + w * h / 4;

  for (a = h; a > 0; a -= 2)
    {
      for (b = w; b > 0; b -= 2)
        {
          *(y++) = *(s++);
          *(u++) = *(s++);
          *(y++) = *(s++);
          *(v++) = *(s++);
        }
      for (b = w; b > 0; b -= 2)
        {
          *(y++) = *(s++);
          s++;
          *(y++) = *(s++);
          s++;
        }
    }
} 


#define SHIFT1  10
#define SHIFT2  12

#define YR24    263
#define YG24    516
#define YB24    100
#define UR24    152
#define UG24    298
#define UB24    450
#define VR24    450
#define VG24    377
#define VB24     73

#define YR15   2104
#define YG15   4130
#define YB15    802
#define UR15   1214
#define UG15   2384
#define UB15   3598
#define VR15   3598
#define VG15   3013
#define VB15    585

#define YR16   2104
#define YG16   2065
#define YB16    802
#define UR16   1214
#define UG16   1192
#define UB16   3598
#define VR16   3598
#define VG16   1506
#define VB16    585


static void
//rgb32_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
//
// Don't know why I need to invert cb and cr in 32 bpp and not in 16 or 15 !!??
//
rgb32_to_yuv420p(unsigned char *lum, unsigned char *cr, unsigned char *cb,
                 unsigned char *src, int width, int height)
{
    int wrap, wrap4, x, y;
    int r, g, b, r1, g1, b1;
    unsigned char *p;

    wrap = width;
    wrap4 = width * 4;
    p = src;
    for(y=0;y<height;y+=2) {
        for(x=0;x<width;x+=2) {
            r = r1 = p[0];
            g = g1 = p[1];
            b = b1 = p[2];
            lum[0] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);
            r = p[4];
            r1 += r;
            g = p[5];
            g1 += g;
            b = p[6];
            b1 += b;
            lum[1] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);

            p += wrap4;
            lum += wrap;

            r = p[0];
            r1 += r;
            g = p[1];
            g1 += g;
            b = p[2];
            b1 += b;
            lum[0] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);
            r = p[4];
            r1 += r;
            g = p[5];
            g1 += g;
            b = p[6];
            b1 += b;
            lum[1] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);
            
            cb[0] = 128 + ((UB24 * b1 - UR24 * r1 - UG24 * g1) >> SHIFT2);
            cr[0] = 128 + ((VR24 * r1 - VG24 * g1 - VB24 * b1) >> SHIFT2);

            cb++;
            cr++;
            p += -wrap4 + 8;
            lum += -wrap + 2;
        }
        p += wrap4;
        lum += wrap;
    }
}

static void
rgb24_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height)
{
    int wrap, wrap3, x, y;
    int r, g, b, r1, g1, b1;
    unsigned char *p;

    wrap = width;
    wrap3 = width * 3;
    p = src;
    for(y=0;y<height;y+=2) {
        for(x=0;x<width;x+=2) {
            r = r1 = p[0];
            g = g1 = p[1];
            b = b1 = p[2];
            lum[0] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);

            r = p[3];
            r1 += r;
            g = p[4];
            g1 += g;
            b = p[5];
            b1 += b;
            lum[1] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);

            p += wrap3;
            lum += wrap;

            r = p[0];
            r1 += r;
            g = p[1];
            g1 += g;
            b = p[2];
            b1 += b;
            lum[0] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);

            r = p[3];
            r1 += r;
            g = p[4];
            g1 += g;
            b = p[5];
            b1 += b;
            lum[1] = 16 + ((YR24 * r + YG24 * g + YB24 * b) >> SHIFT1);
            
            cb[0] = 128 + ((UB24 * b1 - UR24 * r1 - UG24 * g1) >> SHIFT2);
            cr[0] = 128 + ((VR24 * r1 - VG24 * g1 - VB24 * b1) >> SHIFT2);

            cb++;
            cr++;
            p += -wrap3 + 6;
            lum += -wrap + 2;
        }
        p += wrap3;
        lum += wrap;
    }
}

static void
rgb16_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height)
{
    int wrap, x, y;
    int r, g, b, r1, g1, b1;
    unsigned short *p;

    wrap = width;
    p = (unsigned short *)src;
    for(y=0;y<height;y+=2) {
        for(x=0;x<width;x+=2) {
            r = r1 =  p[0] >> 11;
            g = g1 = (p[0] >> 5) & 63;
            b = b1 =  p[0] & 31;
            lum[0] = 16 + ((YR16 * r + YG16 * g + YB16 * b) >> SHIFT1);

            r = p[1] >> 11;
            r1 += r;
            g = (p[1] >> 5) & 63;
            g1 += g;
            b =  p[1] & 31;
            b1 += b;
            lum[1] = 16 + ((YR16 * r + YG16 * g + YB16 * b) >> SHIFT1);

            p += wrap;
            lum += wrap;

            r = p[0] >> 11;
            r1 += r;
            g = (p[0] >> 5) & 63;
            g1 += g;
            b =  p[0] & 31; 
            b1 += b;
            lum[0] = 16 + ((YR16 * r + YG16 * g + YB16 * b) >> SHIFT1);

            r = p[1] >> 11;
            r1 += r;
            g = (p[1] >> 5) & 63;
            g1 += g;
            b =  p[1] & 31; 
            b1 += b;
            lum[1] = 16 + ((YR16 * r + YG16 * g + YB16 * b) >> SHIFT1);

            cb[0] = 128 + ((UB16 * b1 - UR16 * r1 - UG16 * g1) >> SHIFT2);
            cr[0] = 128 + ((VR16 * r1 - VG16 * g1 - VB16 * b1) >> SHIFT2);

            cb++;
            cr++;
            p += -wrap + 2;
            lum += -wrap + 2;
        }
        p += wrap;
        lum += wrap;
    }
}

static void
rgb15_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
                 unsigned char *src, int width, int height)
{
    int wrap, x, y;
    int r, g, b, r1, g1, b1;
    unsigned short *p;

    wrap = width;
    p = (unsigned short *)src;
    for(y=0;y<height;y+=2) {
        for(x=0;x<width;x+=2) {
            r = r1 = (p[0] >> 10) & 31;
            g = g1 = (p[0] >> 5) & 31;
            b = b1 =  p[0] & 31;
            lum[0] = 16 + ((YR15 * r + YG15 * g + YB15 * b) >> SHIFT1);

            r = (p[1] >> 10) & 31;
            r1 += r;
            g = (p[1] >> 5) & 31;
            g1 += g;
            b =  p[1] & 31; 
            b1 += b;
            lum[1] = 16 + ((YR15 * r + YG15 * g + YB15 * b) >> SHIFT1);

            p += wrap;
            lum += wrap;

            r = (p[0] >> 10) & 31;
            r1 += r;
            g = (p[0] >> 5) & 31;
            g1 += g;
            b =  p[0] & 31;
            b1 += b;
            lum[0] = 16 + ((YR15 * r + YG15 * g + YB15 * b) >> SHIFT1);

            r = (p[1] >> 10) & 31;
            r1 += r;
            g = (p[1] >> 5) & 31;
            g1 += g;
            b =  p[1] & 31;
            b1 += b;
            lum[1] = 16 + ((YR15 * r + YG15 * g + YB15 * b) >> SHIFT1);

            cb[0] = 128 + ((UB15 * b1 - UR15 * r1 - UG15 * g1) >> SHIFT2);
            cr[0] = 128 + ((VR15 * r1 - VG15 * g1 - VB15 * b1) >> SHIFT2);

            cb++;
            cr++;
            p += -wrap + 2;
            lum += -wrap + 2;
        }
        p += wrap;
        lum += wrap;
    }
}

static void
rgb32_top_down(unsigned char *dest, unsigned char *foo1, unsigned char *foo2,
               unsigned char *src, int width, int height)
{
  unsigned char *d, *s;
  int y;
  int size;

  d = dest;
  size = 4 * width;
  s = src + size * (height - 1);
  for (y = 0; y < height; y++)
    {
      fast_memcpy(d, s, size);
      d += size;
      s -= size;
    }
}

static void gray_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
			unsigned char *src, int width, int height) {
  int size=width*height;
  fast_memcpy(lum, src, size);
  memset(cb, 128, size/4);
  memset(cr, 128, size/4);
}

static void hi240_to_rgb32_noflip(unsigned char *dest, unsigned char *foo1, unsigned char *foo2,
				  unsigned char *src, int width, int height) {
  static int *rgb32_conv=NULL;
  int i;
  int *dest2=(int *)dest;
  int size=width*height;
  if(rgb32_conv==NULL) {
    rgb32_conv=(int *) malloc(256*sizeof(int));
    for(i=0;i<256;i++) {
      int r=hi240_to_rgb[i][0]*255.0/65535+0.5;
      int g=hi240_to_rgb[i][1]*255.0/65535+0.5;
      int b=hi240_to_rgb[i][2]*255.0/65535+0.5;
      rgb32_conv[i]=(b<<0)|(g<<8)|(r<<16);
    }
  }
  for(i=0;i<size;i++) dest2[i]=rgb32_conv[src[i]];
}

static void hi240_to_rgb32(unsigned char *dest, unsigned char *foo1, unsigned char *foo2,
			   unsigned char *src, int width, int height) {
  static unsigned char *rgb32_buf=NULL;
  rgb32_buf=(unsigned char *) realloc(rgb32_buf,width*height*4);
  hi240_to_rgb32_noflip(rgb32_buf,NULL,NULL,src,width,height);
  rgb32_top_down(dest,NULL,NULL,rgb32_buf,width,height);
}

static void hi240_to_yuv420p(unsigned char *lum, unsigned char *cb, unsigned char *cr,
			     unsigned char *src, int width, int height) {
  static unsigned char *rgb32_buf=NULL;
  rgb32_buf=(unsigned char *) realloc(rgb32_buf,width*height*4);
  hi240_to_rgb32_noflip(rgb32_buf,NULL,NULL,src,width,height);
  rgb32_to_yuv420p(lum,cb,cr,rgb32_buf,width,height);
}


/* function which should be executed in main.c, or by the interface ?*/
void setfps(double fps) 
{
#ifdef SYNCHRO
  divx.fps= 1000 * fps + 0.5;
#endif
}

void write_sub(char *s)
{
  static char *lastsub = NULL;
  static long lastsubbegin;
  long frame;
  int i,j,k,debutligne,lignevide,l,nlignes;
  char *s2;

  if(!subfile) return;
  frame=video_frames;
  if(lastsub!=NULL) {
    if(debug)
      printf("%ld %ld\n",lastsubbegin-lastsubend,
	     frame-lastsubbegin);
    fprintf(subfile,"%ld %ld\n",lastsubbegin-lastsubend,
	    frame-lastsubbegin);
    fputs(lastsub,subfile);
    fputs("\n\n",subfile);
    fflush(subfile);
    lastsubend=frame;
    free(lastsub);
  }
  if(s==NULL || *s=='\0') {lastsub=NULL;return;}
  l=strlen(s);
  s2=(char *)malloc(l+1);

  debutligne=0;
  lignevide=1;
  for(nlignes=0,i=0,j=0;i<l;i++) {
    if(s[i]!='\n' && i!=l-1) {
      if(s[i]!=' ' && s[i]!='#') lignevide=0;
    } else {
      if(s[i]!='\n' && s[i]!='#' && s[i]!=' ') //just case i==l-1
	lignevide=0;
      if(!lignevide && nlignes<8) {
	for(k=debutligne;k<=i;k++,j++)
	  if(s[k]!='#') s2[j]=s[k]; else s2[j]=' ';
	nlignes++;
      }
      debutligne=i+1;
      lignevide=1;
    }
  }
  if(j==0) 
    {free(s2);lastsub=NULL; return;}
  s2[j]='\0';
  lastsub=s2;
  lastsubbegin=frame;
  if(debug)
    printf("lastsub=@@%s@@\n",lastsub);
}



#endif // if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined (HAVE_XVID)
